/*
 *  seed.cc created on Mon, 31 May 1999 21:12:51 -0400 by noe
 */    
#ifndef __SEED_H__
#define __SEED_H__

//STL
#include <vector>
#include <queue>
#include <stack>
#include <list>
//STD
#include <assert.h>
#include <iostream>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <string>
//STR
using namespace std;
#include "macro.h"

   
/**
 * @class seed
 * @brief represents a subset seed 
 */

class seed {
 public:

  /** @name  seed() 
   *  @brief build an empty seed 
   */

  seed(){
    _seedMotif_int = NULL;
    _span = 0;
    _minspan = 0;
    _maxspan = 0;
  }
   
  /** @name  seed(int span) 
   *  @brief build a seed of given span but full of jokers
   *  @param span is the lenght of the seed
   */
 
  seed(int span):_minspan(span),_maxspan(span),_span(span){
    _seedMotif_int  = new int [_span];
    for(int i=0; i<_span; i++)
      _seedMotif_int[i] = 0;

    if (gv_signature_flag){
      checksignature();
      setsignature();
    }

    while(!acceptable())
      next();
  }

  seed(int minspan,int maxspan):_minspan(minspan),_maxspan(maxspan),_span(minspan){
    _seedMotif_int  = new int [_span];
    for(int i=0; i<_span; i++)
      _seedMotif_int[i] = 0;

    if (gv_signature_flag){
      checksignature();
      setsignature();
    }

    while(!acceptable())
      next();
  }




  /** @name  seed(char * desc) 
   *  @brief build a seed according to a descriptor 
   *  @param desc is a string in ['0'-'9''A'-'Z''a'-'z']
   */
  
  //@{
  
  seed (char * desc) {
    /* seed span */
    _span = 0;
    for(char * pc = desc; *pc ; pc++)
      _span++;
    _minspan = _span;
    _maxspan = _span;

    /* seed motif */
    _seedMotif_int  = new int [_span];
    for(int i = 0; i < _span ; i++)
      _seedMotif_int[i] = convert(desc[i]);
  }
  
  
  seed(string & str_desc) {
    const char * desc = str_desc.c_str();
    /* seed span */
    _span = 0;
    for(char * pc = (char *) desc; *pc ; pc++)
      _span++;
    _minspan = _span;
    _maxspan = _span;

    /* seed motif */
    _seedMotif_int  = new int [_span];
    for(int i = 0; i < _span ; i++)
      _seedMotif_int[i] = convert(desc[i]);
  }
  
  //@}

  
  ~seed(){
    if  (_seedMotif_int){
      delete _seedMotif_int;
      _seedMotif_int = NULL;
    }
  };

  /** @name Output functions
   *  @brief these methods are provided to get current seed processed
   */
  
  //@{

  /** @brief gives the integer representation of the seed
   *  @return an int table of lenght span, each element must 
   *          be in the range 0 .. |B|-1
   */
  
  int  *   table()   {return _seedMotif_int;};
  
  /** @brief gives the span of the seed
   *  @return the span of the seed
   */
  int      span()    {return _span;};
  int      minspan() {return _minspan;};
  int      maxspan() {return _maxspan;};
  
        
  /** @brief gives the current seed representation (a string
   *  @brief in ['0'-'9''A'-'Z''a'-'z'])
   *  @return a string object of length span
   */
  string str(){
    string result("");
    for (int i = 0 ; i < _span ; i++) {
      result += rconvert(_seedMotif_int[i]);
    }
    return result;
  }
  
  
  //@}

  
  /** @brief compute the weight of a given seed
   *  @return weight
   */
  double weight(){
    double weight = 0.0;
    for(int i=0; i<_span; i++)
      weight += gv_bsel_weight[_seedMotif_int[i]];
    return weight;
  }
  

  /** @brief check if the current seed is acceptable : correct weight and non 0-weight elements 
   *  @return 1 if the current
   */
  bool acceptable() {
    // FIXME : is is necessary to keep positive weighted elements on the seed borders
    /* 
       if (gv_bsel_weight[_seedMotif_int[0]]       <= (gv_minweight/_span) || 
       gv_bsel_weight[_seedMotif_int[_span-1]] <= (gv_minweight/_span) ) {
       return false;
       }
    */
    double w = weight();
    
    if (w < gv_minweight || w > gv_maxweight)
      return false;
    
    return true;
  }
  


  /** @brief check if the signature is realisable : stop the program otherwise
   */
  
  void checksignature() {
    int sumsignature = 0;
    for(int i = 0 ; i < gv_seed_alphabet_size ; i++)
      sumsignature += gv_signature[i];      
    if (sumsignature > _minspan ){
      ERROR("Invalid Signature","size of the signature is " << sumsignature << " and does not fit inside minimal span " << _minspan);
    }
  }

  /** @brief set the first seed signature
   */

  void setsignature() {
    int pos = 0;
    for(int b = gv_seed_alphabet_size-1 ; b >= 0 && pos < _span ; b--){
      int lettersignature =  gv_signature[b];
      while(lettersignature > 0) { // fill with b
	_seedMotif_int[pos++] = b;
	lettersignature--;
      }
    }
    while (pos < _span) { // fill remaining part with jokers
      _seedMotif_int[pos++] = 0;
    }
  }




  /** @brief generates the next seed 
   *  @return 1 if next seed has been generated, 0 if all the seeds in the set have been generated
   */

  int next() {

    if (gv_signature_flag) {
      /* 1) non trivial signature enumeration */

      for (int b = 1; b < gv_seed_alphabet_size ; b++ ){
	
	int first_b_from_right_to_left = -1;
	int last_bm_before_b           = -1; 
	
	for (int j=_span-1; j >= 0; j--) {
	  if ( _seedMotif_int[j] <  b ) {
	    last_bm_before_b = j; 
	  } else if ( _seedMotif_int[j] == b && last_bm_before_b >= 0 ) {
	    first_b_from_right_to_left = j ;
	    break;
	  }
	}
	if (first_b_from_right_to_left >= 0) {
	  //swap elements
	  int e = _seedMotif_int[first_b_from_right_to_left];
	  _seedMotif_int[first_b_from_right_to_left] =  _seedMotif_int[last_bm_before_b];
	  _seedMotif_int[last_bm_before_b] = e;
	  //reset elements lower than bm
	  for (int b = e-1 ; b > 0 ; b-- ){
	    int last_bm = -1;
	    for(int i = 0 ; i < _span ; i++) {
	      if (_seedMotif_int[i] == b && last_bm >= 0) {
		int e = _seedMotif_int[i];
		_seedMotif_int[i] = _seedMotif_int[last_bm];
		_seedMotif_int[last_bm] = e;
		i = last_bm;
		last_bm = -1;
	      } else {
		if (_seedMotif_int[i] < b && last_bm == -1)
		  last_bm = i;
	      }
	    }
	  }
	  return 1;
	}
      }

      // span changing
      if (_span < _maxspan) {
	_span ++;
	delete _seedMotif_int;
	_seedMotif_int = new int[_span];
	setsignature();
	return 1;
      } else {
	return 0;
      }


    } else {
      
      /* 2) complete enumeration */
    start:
      int j=_span-1;
      
      while(_seedMotif_int[j] == (gv_seed_alphabet_size-1)){
	j--;
	if (j < 0) {
	  if (_span < _maxspan) {
	    _span ++;
	    delete _seedMotif_int;
	    _seedMotif_int = new int[_span];
	    goto reset;
	  } else {
	    return 0;
	  }
	}
      }
      
      _seedMotif_int[j]++;
    reset:
      j++;
      for(int i = j ; i < _span ; i++) {
	_seedMotif_int[i] = 0;
      }
      
      while(!acceptable())
	goto start;
      return 1;
    }/* 2) */
    
  }
  
  
  
  
  /** @brief generates a random seed of given span 
   *  @return 1 all the time
   */
  int random(){
    if (gv_signature_flag) {


      /* 1) signature random generation */
      // check signature
      int sumsignature = 0;
      for(int i = 0 ; i < gv_seed_alphabet_size ; i++)
	sumsignature += gv_signature[i];      
      if (sumsignature > _minspan ){
	ERROR("Invalid Signature","size of the signature is " << sumsignature << " and does not fit inside minimal span " << _minspan);
      }
      
      // select a new span 
      int newspan = rand()%(_maxspan - _minspan + 1) + _minspan;
      if (newspan != _span) {
	_span = newspan;
	delete _seedMotif_int;
	_seedMotif_int = new int[_span];
      }
       
      
      // "fill and swap" randomization : 

      // fill
      setsignature();      
      // swap
    swap:
      for(int i= 0 ; i < _span ; i++){
	int j = rand()%_span;
	int letter_tmp    = _seedMotif_int[i];
	assert(letter_tmp >= 0);
	assert(letter_tmp <  gv_seed_alphabet_size);
	
	_seedMotif_int[i] = _seedMotif_int[j];
	_seedMotif_int[j] = letter_tmp;
      }
      
     
      while(!acceptable())
	goto swap; /* FIXME  can loop infinitely if a bad signature is given */

    } else {
      /* 2) complete enumeration */
      int newspan = rand()%(_maxspan - _minspan + 1) + _minspan;
      if (newspan != _span) {
	_span = newspan;
	delete _seedMotif_int;
	_seedMotif_int = new int[_span];
      }
      for(int i = 0 ; i < _span ; i++) {
	_seedMotif_int[i] = rand() % gv_seed_alphabet_size;
      }
      
      while(!acceptable())
	next();
      
    }
    return 1;
  }



  
  /** @brief ouput operator
   *  @return 
   */
  
  friend ostream& operator<<(ostream& os, seed& s);

 

 protected:
  /// Current seed motif represented as an int array {0,1,2, ...} of lenght "_span"
  int  * _seedMotif_int;
  /// Fixed seed span 
  int    _minspan,_maxspan,_span;
  /// convert a char into a number
  /// '0' -> 0, '1' -> 1 ... '9' -> 9, 'A' -> 10 ... 'Z' -> 35, 'a' -> 36 ...
  int convert(char c){
    if (c >= '0' && c <= '9') {
      if (c - '0' >= gv_seed_alphabet_size) {
	cerr << "* Error : out of bound char " << c << " inside the seed descriptor" << endl;
	exit(-1);
      }
      return c - '0';
    } else if (c >= 'A' && c <= 'Z') {
      if (c - 'A' + 10 >= gv_seed_alphabet_size) {
	cerr << "* Error : out of bound char " << c << " inside the seed descriptor" << endl;
	exit(-1);
      }
      return c - 'A' + 10;
    } else if (c >= 'a' && c <= '~') {
      if (c - 'a' + 36 >= gv_seed_alphabet_size) {
	cerr << "* Error : out of bound char " << c << " inside the seed descriptor" << endl;
	exit(-1);
      } 
      return c - 'a' + 36;
    } else if (c >= '!') {
      return c - '!' + 66;
    } else {
      return 0;
    }
    return 0;
  }
  
  // convert a number into a char
  char rconvert(int i){
    if (i < 10)
      return (char)i + '0';
    else 
      if (i < 36)
	return (char)i - 10 + 'A';
      else
	if (i < 66)
	  return (char)i - 36 + 'a';
	else
	  return (char)i - 66 + '!'; 
  }
 
};


#endif

